# Predefined physical constants

from math import pi

QE = -1.60217662e-19  # electron charge, [C]
ME = 9.10938356e-31  # electron mass, [kg]
QI = 1.60217662e-19  # proton mass, [C]
MI = 1.673557546e-27  # proton mass, [kg]
C = 299792458.0  # speed of light, [m/s]
MU0 = 4 * pi * 1e-7  # Vacuum permeability, [H/m]
EPSILON0 = 1 / (C**2 * MU0)  # Vacuum permittivity, [F/m]
KB = 1.38064852e-23  # Boltzmann constant, [m²kg/(s²K)]
RE = 6.371e6  # Earth radius, [m]

speciesdict = {
    "avgs": "p",
    "proton": "p",
    "helium": "He",
    "oxygen": "O",
    "electron": "e",
}
speciesamu = {
    "avgs": 1,
    "proton": 1,
    "helium": 4,
    "oxygen": 16,
    "electron": 5.4461702e-4,
}
speciescharge = {
    "avgs": 1,
    "proton": 1,
    "helium": 2,
    "oxygen": 1,
    "electron": -1,
}

# Define units, LaTeX markup names, and LaTeX markup units for intrinsic values
units_predefined = {
    "Rhom": ("kg/m3", r"$\rho_m$", r"$\mathrm{kg}\,\mathrm{m}^{-3}$"),
    "rhoq": ("C/m3", r"$\rho_q$", r"$\mathrm{C}\,\mathrm{m}^{-3}$"),
    "rho": ("1/m3", r"$n_\mathrm{p}$", r"$\mathrm{m}^{-3}$"),
    "rho_v": ("1/m2s", r"$\Gamma_\mathrm{p}$", r"$\mathrm{m}^{-2}$s"),
    "v": ("m/s", r"$V$", r"$\mathrm{m}\,\mathrm{s}^{-1}$"),
    "b": ("T", r"$B$", "T"),
    "b_vor": ("T", r"$B_\mathrm{vol}$", "T"),
    "background_b": ("T", r"$B_\mathrm{bg}$", "T"),
    "perturbed_b": ("T", r"$B_\mathrm{pert}$", "T"),
    "bgb": ("T", r"$B_\mathrm{bg}$", "T"),
    "perb": ("T", r"B_\mathrm{pert}$", "T"),
    "perb_vor": ("T", r"B_\mathrm{vol,pert}$", "T"),
    "e": ("V/m", r"$E$", r"$\mathrm{V}\,\mathrm{m}^{-1}$"),
    "e_vor": ("V/m", r"$E_\mathrm{vol}$", r"$\mathrm{V}\,\mathrm{m}^{-1}$"),
    "exhall_000_100": (
        "V/m",
        r"$E_\mathrm{Hall,000,100}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "exhall_001_101": (
        "V/m",
        r"$E_\mathrm{Hall,001,101}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "exhall_010_110": (
        "V/m",
        r"$E_\mathrm{Hall,010,110}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "exhall_011_111": (
        "V/m",
        r"$E_\mathrm{Hall,011,111}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "eyhall_000_010": (
        "V/m",
        r"$E_\mathrm{Hall,000,010}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "eyhall_001_011": (
        "V/m",
        r"$E_\mathrm{Hall,001,011}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "eyhall_100_110": (
        "V/m",
        r"$E_\mathrm{Hall,100,110}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "eyhall_101_111": (
        "V/m",
        r"$E_\mathrm{Hall,101,111}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "ezhall_000_001": (
        "V/m",
        r"$E_\mathrm{Hall,000,001}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "ezhall_010_011": (
        "V/m",
        r"$E_\mathrm{Hall,010,011}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "ezhall_100_101": (
        "V/m",
        r"$E_\mathrm{Hall,100,101}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "ezhall_110_111": (
        "V/m",
        r"$E_\mathrm{Hall,110,111}$",
        r"$\mathrm{V}\,\mathrm{m}^{-1}$",
    ),
    "pressure": ("Pa", r"$P$", "Pa"),
    "pressure_dt2": ("Pa", r"$P_{\mathrm{d}t/2}}$", "Pa"),
    "pressure_r": ("Pa", r"$P_r$", "Pa"),
    "pressure_v": ("Pa", r"$P_v$", "Pa"),
    "ptensordiagonar": ("Pa", r"$\mathcal{P}_\mathrm{diag}$", "Pa"),
    "ptensoroffdiagonar": ("Pa", r"$\mathcal{P}_\mathrm{off-diag}$", "Pa"),
    "minvalue": ("s3/m6", r"$f_\mathrm{Min}$", r"$\mathrm{m}^{-6}\,\mathrm{s}^{3}$"),
    "effectivesparsitythreshold": (
        "s3/m6",
        r"$f_\mathrm{Min}$",
        r"$\mathrm{m}^{-6}\,\mathrm{s}^{3}$",
    ),
    "rho_loss_adjust": (
        "1/m3",
        r"$\Delta_\mathrm{loss} n_\mathrm{p}$",
        r"$\mathrm{m}^{-3}$",
    ),
    "energydensity": (
        "eV/cm3",
        r"$\rho_{\mathrm{energy}}$",
        r"$\mathrm{eV}\,\mathrm{cm}^{-3}$",
    ),
    "precipitationdiffflux": (
        "1/(cm2 sr s eV)",
        r"$\Delta F_\mathrm{precipitation}$",
        r"$\mathrm{cm}^{-2} \,\mathrm{sr}^{-1}\,\mathrm{s}^{-1}\,\mathrm{eV}^{-1}$",
    ),
    "T": ("K", r"$T$", "K"),
    "Tpar": ("K", r"$T$", "K"),
    "Tperp": ("K", r"$T$", "K"),
    "Panisotropy": ("", r"$P_\perp / P_\parallel$", ""),
    "Tanisotropy": ("", r"$T_\perp / T_\parallel$", ""),
    "VS": ("m/s", r"$V_S$", r"$\mathrm{m}\,\mathrm{s}^{-1}$"),
    "VA": ("m/s", r"$V_A$", r"$\mathrm{m}\,\mathrm{s}^{-1}$"),
    "MS": ("", r"$M_S$", ""),
    "MA": ("", r"$M_A$", ""),
    "Ppar": ("Pa", r"$P_\parallel$", "Pa"),
    "Pperp": ("Pa", r"$P_\perp$", "Pa"),
    "Beta": ("", r"$\beta$", ""),
    "BetaStar": ("", r"$\beta^\ast$", ""),
    "Gyroperiod": ("s", r"$T_{gyro}$", "s"),
    "PlasmaPeriod": ("s", r"$T_{plasma}$", "s"),
    "Gyrofrequency": ("rad/s", r"$\omega_{g}$", r"\mathrm{rad}/\mathrm{s}"),
    "Omegap": ("rad/s", r"$\omega_{p}$", r"\mathrm{rad}/\mathrm{s}"),
}


def pass_op(variable):
    # do nothing
    return variable


def magnitude(variable):
    return np.linalg.norm(np.asarray(variable), axis=-1)


def sumv(variable):
    # Note: this is used to sum over multipops, thus the summing axis is zero
    if np.ndim(variable) > 3:
        print("Error: Number of dimensions is too large")
        return
    else:
        # First dimension: populations
        # Second dimension: cells
        # Third dimension: components
        return np.sum(np.array(variable), axis=0)


# Dict of operators. The user can apply these to any variable,
# including more general datareducers. Can only be used to reduce one
# variable at a time
data_operators = {}
data_operators["pass"] = pass_op
data_operators["magnitude"] = magnitude
data_operators["sum"] = sumv



from __future__ import annotations

import numpy as np
import math
import warnings
from typing import Callable
from collections import namedtuple
from enum import Enum
from pyvlasiator.vlsv import Vlsv
from pyvlasiator.vlsv.reader import _getdim2d
from pyvlasiator.vlsv.variables import RE


class ColorScale(Enum):
    """
    Represents the available color scales for data visualization.

    Attributes:
        - Linear (1): A linear color scale, where colors are evenly distributed across the data range.
        - Log (2): A logarithmic color scale, suitable for data with a wide range of values, where smaller values are more emphasized.
        - SymLog (3): A symmetric logarithmic color scale, similar to Log but with symmetry around zero, useful for data with both positive and negative values.
    """

    Linear = 1
    Log = 2
    SymLog = 3


class AxisUnit(Enum):
    """
    Specifies the units for representing values on an axis.

    Attributes:
        - EARTH (1): Units based on Earth's physical properties, such as Earth's radius for distance and km/s for velocity.
        - SI (2): Units from the International System of Units (SI), such as meters, seconds, kilograms, etc.
    """

    EARTH = 1
    SI = 2


# Plotting arguments
PlotArgs = namedtuple(
    "PlotArgs",
    [
        "axisunit",
        "sizes",
        "plotrange",
        "origin",
        "idlist",
        "indexlist",
        "str_title",
        "strx",
        "stry",
        "cb_title",
    ],
)


def plot(
    self: Vlsv,
    var: str = "",
    ax=None,
    figsize: tuple[float, float] | None = None,
    **kwargs,
) -> list[matplotlib.lines.Line2D]:
    """
    Plots 1D data from a VLSV file.

    Parameters
    ----------
    var : str
        Name of the variable to plot from the VLSV file.
    ax : matplotlib.axes._axes.Axes, optional
        Axes object to plot on. If not provided, a new figure and axes will be created.
    figsize : tuple[float, float], optional
        Size of the figure in inches (width, height). Only used if a new
        figure is created.
    **kwargs
        Additional keyword arguments passed to Matplotlib's `plot` function.

    Returns
    -------
    list[matplotlib.lines.Line2D]
        A list of created Line2D objects.

    Raises
    ------
    ValueError
        If the specified variable is not found in the VLSV file.

    Examples
    --------
    >>> vlsv_file = Vlsv("my_vlsv_file.vlsv")
    >>> axes = vlsv_file.plot("proton/vg_rho")  # Plot density on a new figure
    >>> axes = vlsv_file.plot("helium/vg_v", ax=my_axes)  # Plot velocity on an existing axes
    """

    import matplotlib.pyplot as plt

    fig = kwargs.pop(
        "figure", plt.gcf() if plt.get_fignums() else plt.figure(figsize=figsize)
    )
    if ax is None:
        ax = fig.gca()

    if not self.has_variable(var):
        raise ValueError(f"Variable {var} not found in the file")

    x = np.linspace(self.coordmin[0], self.coordmax[0], self.ncells[0])

    data = self.read_variable(var)
    axes = ax.plot(x, data, **kwargs)

    return axes


def pcolormesh(
    self: Vlsv,
    var: str = "",
    axisunit: AxisUnit = AxisUnit.EARTH,
    colorscale: ColorScale = ColorScale.Linear,
    addcolorbar: bool = True,
    vmin: float = float("-inf"),
    vmax: float = float("inf"),
    extent: list = [0.0, 0.0, 0.0, 0.0],
    comp: int = -1,
    ax=None,
    figsize: tuple[float, float] | None = None,
    **kwargs,
):
    """
    Plots 2D VLSV data using pcolormesh.

    Parameters
    ----------
    var : str
        Name of the variable to plot from the VLSV file.
    axisunit : AxisUnit
        Unit of the axis, `AxisUnit.EARTH` or `AxisUnit.SI`.
    addcolorbar : bool
        Add colorbar to the right.
    colorscale: ColorScale
        Color scale of the data, `ColorScale.Linear`, `ColorScale.Log`, or `ColorScale.SymLog`.
    extent : list
        Extent of the domain (WIP).
    comp : int
        Vector composition of data, -1 is magnitude, 0 is x, 1 is y, and 2 is z.
    ax : matplotlib.axes._axes.Axes, optional
        Axes object to plot on. If not provided, a new figure and axes
        will be created using `set_figure`.
    figsize : tuple[float, float], optional
        Size of the figure in inches. Only used if a new figure is created.
    **kwargs
        Additional keyword arguments passed to `ax.pcolormesh`.

    Returns
    -------
    matplotlib.figure.Figure
        The created or existing figure object.

    Raises
    ------
    ValueError
        If the specified variable is not found in the VLSV file.

    Examples
    --------
    >>> vlsv_file = Vlsv("my_vlsv_file.vlsv")
    >>> # Plot density on a new figure
    >>> fig = vlsv_file.pcolormesh("proton/vg_rho")
    >>> # Plot velocity on an existing axes
    >>> ax = ...  # Existing axes object
    >>> fig = vlsv_file.pcolormesh("proton/vg_v", ax=ax)
    """

    fig, ax = set_figure(ax, figsize, **kwargs)

    return _plot2d(
        self,
        ax.pcolormesh,
        var=var,
        ax=ax,
        comp=comp,
        axisunit=axisunit,
        colorscale=colorscale,
        addcolorbar=addcolorbar,
        vmin=vmin,
        vmax=vmax,
        extent=extent,
        **kwargs,
    )


def contourf(
    self: Vlsv,
    var: str = "",
    axisunit: AxisUnit = AxisUnit.EARTH,
    colorscale: ColorScale = ColorScale.Linear,
    addcolorbar: bool = True,
    vmin: float = float("-inf"),
    vmax: float = float("inf"),
    extent: list = [0.0, 0.0, 0.0, 0.0],
    comp: int = -1,
    ax=None,
    figsize: tuple[float, float] | None = None,
    **kwargs,
):
    fig, ax = set_figure(ax, figsize, **kwargs)

    return _plot2d(
        self,
        ax.contourf,
        var=var,
        ax=ax,
        comp=comp,
        axisunit=axisunit,
        colorscale=colorscale,
        addcolorbar=addcolorbar,
        vmin=vmin,
        vmax=vmax,
        extent=extent,
        **kwargs,
    )


def contour(
    self: Vlsv,
    var: str = "",
    axisunit: AxisUnit = AxisUnit.EARTH,
    colorscale: ColorScale = ColorScale.Linear,
    addcolorbar: bool = True,
    vmin: float = float("-inf"),
    vmax: float = float("inf"),
    extent: list = [0.0, 0.0, 0.0, 0.0],
    comp: int = -1,
    ax=None,
    figsize: tuple[float, float] | None = None,
    **kwargs,
):
    fig, ax = set_figure(ax, figsize, **kwargs)

    return _plot2d(
        self,
        ax.contour,
        var=var,
        ax=ax,
        comp=comp,
        axisunit=axisunit,
        colorscale=colorscale,
        addcolorbar=addcolorbar,
        vmin=vmin,
        vmax=vmax,
        extent=extent,
        **kwargs,
    )


def _plot2d(
    meta: Vlsv,
    plot_func: Callable,
    var: str = "",
    axisunit: AxisUnit = AxisUnit.EARTH,
    colorscale: ColorScale = ColorScale.Linear,
    addcolorbar: bool = True,
    vmin: float = float("-inf"),
    vmax: float = float("inf"),
    extent: list = [0.0, 0.0, 0.0, 0.0],
    comp: int = -1,
    ax=None,
    **kwargs,
):
    """
    Plot 2d data.

    Parameters
    ----------
    var : str
        Variable name from the VLSV file.

    Returns
    -------

    """

    if not meta.has_variable(var):
        raise ValueError(f"Variable {var} not found in the file")

    if meta.ndims() == 3 or meta.maxamr > 0:
        # check if origin and normal exist in kwargs
        normal = kwargs["normal"] if "normal" in kwargs else 1
        origin = kwargs["origin"] if "origin" in kwargs else 0.0
        kwargs.pop("normal", None)
        kwargs.pop("origin", None)

        pArgs = set_args(meta, var, axisunit, normal, origin)
        data = prep2dslice(meta, var, normal, comp, pArgs)
    else:
        pArgs = set_args(meta, var, axisunit)
        data = prep2d(meta, var, comp)

    x1, x2 = get_axis(pArgs.axisunit, pArgs.plotrange, pArgs.sizes)

    if var in ("fg_b", "fg_e", "vg_b_vol", "vg_e_vol") or var.endswith("vg_v"):
        _fillinnerBC(data)

    norm, ticks = set_colorbar(colorscale, vmin, vmax, data)

    range1 = range(
        np.searchsorted(x1, extent[0]), np.searchsorted(x1, extent[1], side="right")
    )
    range2 = range(
        np.searchsorted(x2, extent[2]), np.searchsorted(x2, extent[3], side="right")
    )

    c = plot_func(x1, x2, data, **kwargs)

    configure_plot(c, ax, pArgs, ticks, addcolorbar)

    return c


def streamplot(
    meta: Vlsv,
    var: str,
    ax=None,
    comp: str = "xy",
    axisunit: AxisUnit = AxisUnit.EARTH,
    origin: float = 0.0,
    **kwargs,
) -> matplotlib.streamplot.StreamplotSet:
    """
    Creates a streamplot visualization of a vector field from a VLSV dataset.

    Parameters
    ----------
    meta: Vlsv
        A VLSV metadata object containing the dataset information.
    var: str
        The name of the vector variable to visualize.
    ax: matplotlib.axes.Axes, optional
        The axes object to plot on. If not provided, a new figure and axes
        will be created.
    comp: str, optional
        The components of the vector to plot, specified as a string containing
        "x", "y", or "z" (e.g., "xy" for x-y components). Defaults to "xy".
    axisunit: AxisUnit, optional
        The unit system for the plot axes. Defaults to AxisUnit.EARTH.
    origin: float, optional
        The origin point for slice plots. Defaults to 0.0.
    **kwargs
        Additional keyword arguments passed to Matplotlib's streamplot function.

    Returns
    -------
    matplotlib.streamplot.StreamplotSet
        The streamplot object created by Matplotlib.
    """

    X, Y, v1, v2 = set_vector(meta, var, comp, axisunit, origin)
    fig, ax = set_figure(ax, **kwargs)

    s = ax.streamplot(X, Y, v1, v2, **kwargs)

    return s


def set_vector(
    meta: Vlsv, var: str, comp: str, axisunit: AxisUnit, origin: float = 0.0
):
    """
    Extracts and prepares vector data for plotting from a VLSV dataset.

    Parameters
    ----------
    meta: Vlsv
        A VLSV metadata object containing the dataset information.
    var: str
        The name of the vector variable to extract.
    comp: str
        The components of the vector to extract, specified as a string
        containing "x", "y", or "z" (e.g., "xy" for x-y components).
    axisunit: AxisUnit
        The unit system for the plot axes.
    origin: float, optional
        The origin point for slice plots. Defaults to 0.0.

    Returns
    -------
    tuple
        A tuple containing:
            - x: The x-axis coordinates for plotting.
            - y: The y-axis coordinates for plotting.
            - v1: The values of the first vector component.
            - v2: The values of the second vector component.

    Raises
    ------
    ValueError
        If the specified variable is not a vector variable.
    """

    ncells = meta.ncells
    maxamr = meta.maxamr
    coordmin, coordmax = meta.coordmin, meta.coordmax

    if "x" in comp:
        v1_ = 0
        if "y" in comp:
            dir = 2
            v2_ = 1
            sizes = _getdim2d(ncells, maxamr, 2)
            plotrange = (coordmin[0], coordmax[0], coordmin[1], coordmax[1])
        else:
            dir = 1
            v2_ = 2
            sizes = _getdim2d(ncells, maxamr, 1)
            plotrange = (coordmin[0], coordmax[0], coordmin[2], coordmax[2])
    else:
        dir = 0
        v1_, v2_ = 1, 2
        sizes = _getdim2d(ncells, maxamr, 0)
        plotrange = (coordmin[1], coordmax[1], coordmin[2], coordmax[2])

    data = meta.read_variable(var)

    if not var.startswith("fg_"):  # vlasov grid
        if data.ndim != 2 and data.shape[0] == 3:
            raise ValueError("Vector variable required!")
        if meta.maxamr == 0:
            data = data.reshape((sizes[1], sizes[0], 3))
            v1 = data[:, :, v1_]
            v2 = data[:, :, v2_]
        else:
            sliceoffset = origin - coordmin[dir]
            idlist, indexlist = meta.getslicecell(
                sliceoffset, dir, coordmin[dir], coordmax[dir]
            )
            v2D = data[indexlist, :]
            v1 = meta.refineslice(idlist, v2D[:, v1_], dir)
            v2 = meta.refineslice(idlist, v2D[:, v2_], dir)
    else:  # FS grid
        data = np.squeeze(data)
        v1 = np.transpose(data[:, :, v1_])
        v2 = np.transpose(data[:, :, v2_])

    x, y = get_axis(axisunit, plotrange, sizes)

    return x, y, v1, v2


def set_figure(ax, figsize: tuple = (10, 6), **kwargs) -> tuple:
    """
    Sets up a Matplotlib figure and axes for plotting.

    Parameters
    ----------
    ax: matplotlib.axes.Axes, optional
        An existing axes object to use for plotting. If not provided, a new
        figure and axes will be created.
    figsize: tuple, optional
        The desired figure size in inches, as a tuple (width, height).
        Defaults to (10, 6).
    **kwargs
        Additional keyword arguments passed to Matplotlib's figure() function
        if a new figure is created.

    Returns
    -------
    tuple
        A tuple containing:
            - fig: The Matplotlib figure object.
            - ax: The Matplotlib axes object.
    """

    import matplotlib.pyplot as plt

    fig = kwargs.pop(
        "figure", plt.gcf() if plt.get_fignums() else plt.figure(figsize=figsize)
    )
    if ax is None:
        ax = fig.gca()

    return fig, ax


def set_args(
    meta: Vlsv,
    var: str,
    axisunit: AxisUnit = AxisUnit.EARTH,
    dir: int = -1,
    origin: float = 0.0,
) -> PlotArgs:
    """
    Set plot-related arguments of `var` in `axisunit`.

    Parameters
    ----------
    var : str
        Variable name from the VLSV file.
    axisunit : AxisUnit
        Unit of the axis.
    dir : int
        Normal direction of the 2D slice, 0 for x, 1 for y, and 2 for z.
    origin : float
        Origin of the 2D slice.

    Returns
    -------
    PlotArgs

    See Also
    --------
    :func:`pcolormesh`
    """

    ncells, coordmin, coordmax = meta.ncells, meta.coordmin, meta.coordmax

    if dir == 0:
        seq = (1, 2)
    elif dir == 1 or (ncells[1] == 1 and ncells[2] != 1):  # polar
        seq = (0, 2)
        dir = 1
    elif dir == 2 or (ncells[2] == 1 and ncells[1] != 1):  # ecliptic
        seq = (0, 1)
        dir = 2
    else:
        raise ValueError("1D data detected. Please use 1D plot functions.")

    plotrange = (coordmin[seq[0]], coordmax[seq[0]], coordmin[seq[1]], coordmax[seq[1]])
    axislabels = tuple(("X", "Y", "Z")[i] for i in seq)
    # Scale the sizes to the highest refinement level for data to be refined later
    sizes = tuple(ncells[i] << meta.maxamr for i in seq)

    if dir == -1:
        idlist, indexlist = np.empty(0, dtype=int), np.empty(0, dtype=int)
    else:
        sliceoffset = origin - coordmin[dir]
        idlist, indexlist = meta.getslicecell(
            sliceoffset, dir, coordmin[dir], coordmax[dir]
        )

    if axisunit == AxisUnit.EARTH:
        unitstr = r"$R_E$"
    else:
        unitstr = r"$m$"
    strx = axislabels[0] + " [" + unitstr + "]"
    stry = axislabels[1] + " [" + unitstr + "]"

    str_title = f"t={meta.time:4.1f}s"

    datainfo = meta.read_variable_meta(var)

    if not datainfo.variableLaTeX:
        cb_title = datainfo.variableLaTeX + " [" + datainfo.unitLaTeX + "]"
    else:
        cb_title = ""

    return PlotArgs(
        axisunit,
        sizes,
        plotrange,
        origin,
        idlist,
        indexlist,
        str_title,
        strx,
        stry,
        cb_title,
    )


def prep2d(meta: Vlsv, var: str, comp: int = -1):
    """
    Obtain data of `var` for 2D plotting. Use `comp` to select vector components.

    Parameters
        ----------
        meta : Vlsv
            Metadata corresponding to the file.
        var : str
            Name of the variable.
        comp : int
            Vector component. -1 refers to the magnitude of the vector.
        Returns
        -------
        numpy.ndarray
    """

    dataRaw = _getdata2d(meta, var)

    if dataRaw.ndim == 3:
        if comp != -1:
            data = dataRaw[:, :, comp]
        else:
            data = np.linalg.norm(dataRaw, axis=2)
        if var.startswith("fg_"):
            data = np.transpose(data)
    else:
        data = dataRaw

    return data


def prep2dslice(meta: Vlsv, var: str, dir: int, comp: int, pArgs: PlotArgs):
    origin = pArgs.origin
    idlist = pArgs.idlist
    indexlist = pArgs.indexlist

    data3D = meta.read_variable(var)

    if var.startswith("fg_") or data3D.ndim > 2:  # field or derived quantities, fsgrid
        ncells = meta.ncells * 2**meta.maxamr
        if not dir in (0, 1, 2):
            raise ValueError(f"Unknown normal direction {dir}")

        sliceratio = (origin - meta.coordmin[dir]) / (
            meta.coordmax[dir] - meta.coordmin[dir]
        )
        if not (0.0 <= sliceratio <= 1.0):
            raise ValueError("slice plane index out of bound!")
        # Find the cut plane index for each refinement level
        icut = int(np.floor(sliceratio * ncells[dir]))
        if dir == 0:
            if comp != -1:
                data = data3D[icut, :, :, comp]
            else:
                data = np.linalg.norm(data3D[icut, :, :, :], axis=3)
        elif dir == 1:
            if comp != -1:
                data = data3D[:, icut, :, comp]
            else:
                data = np.linalg.norm(data3D[:, icut, :, :], axis=3)
        elif dir == 2:
            if comp != -1:
                data = data3D[:, :, icut, comp]
            else:
                data = np.linalg.norm(data3D[:, :, icut, :], axis=3)
    else:  # moments, dccrg grid
        # vlasov grid, AMR
        if data3D.ndim == 1:
            data2D = data3D[indexlist]

            data = meta.refineslice(idlist, data2D, dir)
        elif data3D.ndim == 2:
            data2D = data3D[indexlist, :]

            if comp in (0, 1, 2):
                slice = data2D[:, comp]
                data = meta.refineslice(idlist, slice, dir)
            elif comp == -1:
                datax = meta.refineslice(idlist, data2D[:, 0], dir)
                datay = meta.refineslice(idlist, data2D[:, 1], dir)
                dataz = meta.refineslice(idlist, data2D[:, 2], dir)
                data = np.fromiter(
                    (np.linalg.norm([x, y, z]) for x, y, z in zip(datax, datay, dataz)),
                    dtype=float,
                )
            else:
                slice = data2D[:, comp]
                data = meta.refineslice(idlist, slice, dir)

    return data


def _getdata2d(meta: Vlsv, var: str) -> np.ndarray:
    """
    Retrieves and reshapes 2D data from a VLSV dataset.

    Raises:
        ValueError: If the dataset is not 2D.
    """

    if meta.ndims() != 2:
        raise ValueError("2D outputs required")
    sizes = [i for i in meta.ncells if i != 1]
    data = meta.read_variable(var)
    if data.ndim in (1, 3) or data.shape[-1] == 1:
        data = data.reshape((sizes[1], sizes[0]))
    elif var.startswith("fg_"):
        data = data.reshape((sizes[0], sizes[1], 3))
    else:
        data = data.reshape((sizes[1], sizes[0], 3))

    return data


def get_axis(axisunit: AxisUnit, plotrange: tuple, sizes: tuple) -> tuple:
    """
    Generates the 2D domain axis coordinates, potentially applying Earth radius scaling.

    Parameters
    ----------
    axisunit: AxisUnit
        The unit system for the plot axes.
    plotrange: tuple
        A tuple containing the minimum and maximum values for both axes (xmin, xmax, ymin, ymax).
    sizes: tuple
        A tuple containing the number of points for each axis (nx, ny).

    Returns
    -------
    tuple
        A tuple containing the x and y axis coordinates.
    """

    scale_factor = 1.0 / RE if axisunit == AxisUnit.EARTH else 1.0
    start = tuple(s * scale_factor for s in plotrange[:2])
    stop = tuple(s * scale_factor for s in plotrange[2:])

    # Vectorized generation of coordinates for efficiency
    x, y = np.linspace(*start, num=sizes[0]), np.linspace(*stop, num=sizes[1])

    return x, y


def _fillinnerBC(data: np.ndarray):
    """
    Fill sparsity/inner boundary cells with NaN.
    """
    data[data == 0] = np.nan


def set_colorbar(
    colorscale: ColorScale = ColorScale.Linear,
    v1: float = np.nan,
    v2: float = np.nan,
    data: np.ndarray = np.array([1.0]),
    linthresh: float = 1.0,
    logstep: float = 1.0,
    linscale: float = 0.03,
):
    """
    Creates a color normalization object and tick values for a colorbar.

    Parameters
    ----------
    colorscale: ColorScale, optional
        The type of color scale to use. Can be 'Linear', 'Log', or 'SymLog'.
        Defaults to 'Linear'.
    v1: float, optional
        The minimum value for the colorbar. Defaults to np.nan, which means
        it will be inferred from the data.
    v2: float, optional
        The maximum value for the colorbar. Defaults to np.nan, which means
        it will be inferred from the data.
    data: np.ndarray, optional
        The data to use for inferring the colorbar limits if v1 and v2 are
        not provided. Defaults to np.array([1.0]).
    linthresh: float, optional
        The threshold value for symmetric log color scales. Defaults to 1.0.
    logstep: float, optional
        The step size for tick values in log color scales. Defaults to 1.0.
    linscale: float, optional
        A scaling factor for linear regions in symmetric log color scales.
        Defaults to 0.03.

    Returns
    -------
    tuple
        A tuple containing:
            - norm: A Matplotlib color normalization object for the colorbar.
            - ticks: A list of tick values for the colorbar.

    Raises
    ------
    ValueError
        If an invalid colorscale type is provided.

    Notes
    -----
    - The 'SymLog' colorscale is currently not fully implemented.
    """
    import matplotlib

    vmin, vmax = set_plot_limits(v1, v2, data, colorscale)
    if colorscale == ColorScale.Linear:
        levels = matplotlib.ticker.MaxNLocator(nbins=255).tick_values(vmin, vmax)
        norm = matplotlib.colors.BoundaryNorm(levels, ncolors=256, clip=True)
        ticks = matplotlib.ticker.LinearLocator(numticks=9)
    elif colorscale == ColorScale.Log:  # logarithmic
        norm = matplotlib.colors.LogNorm(vmin, vmax)
        ticks = matplotlib.ticker.LogLocator(base=10, subs=range(0, 9))
    else:  # symmetric log
        logthresh = int(math.floor(math.log10(linthresh)))
        minlog = int(math.ceil(math.log10(-vmin)))
        maxlog = int(math.ceil(math.log10(vmax)))
        # TODO: fix this!
        # norm = matplotlib.colors.SymLogNorm(linthresh, linscale, vmin, vmax, base=10)
        # ticks = [ [-(10.0**x) for x in minlog:-logstep:logthresh]..., 0.0,
        #   [10.0**x for x in logthresh:logstep:maxlog]..., ]

    return norm, ticks


def set_plot_limits(
    vmin: float,
    vmax: float,
    data: np.ndarray,
    colorscale: ColorScale = ColorScale.Linear,
) -> tuple:
    """
    Calculates appropriate plot limits based on data and colorscale.
    """

    if colorscale in (ColorScale.Linear, ColorScale.SymLog):
        vmin = vmin if not math.isinf(vmin) else np.nanmin(data)
        vmax = vmax if not math.isinf(vmax) else np.nanmax(data)
    else:  # Logarithmic colorscale
        positive_data = data[data > 0.0]  # Exclude non-positive values
        vmin = vmin if not math.isinf(vmin) else np.min(positive_data)
        vmax = vmax if not math.isinf(vmax) else np.max(positive_data)

    return vmin, vmax


def configure_plot(
    c: matplotlib.cm.ScalarMappable,  # Assuming c is a colormap or similar
    ax: matplotlib.pyplot.Axes,
    plot_args: PlotArgs,
    ticks: list,
    add_colorbar: bool = True,
):
    """
    Configures plot elements based on provided arguments.
    """

    import matplotlib.pyplot as plt

    title = plot_args.str_title
    x_label = plot_args.strx
    y_label = plot_args.stry
    colorbar_title = plot_args.cb_title

    # Add colorbar if requested
    if add_colorbar:
        cb = plt.colorbar(c, ax=ax, ticks=ticks, fraction=0.04, pad=0.02)
        if colorbar_title:
            cb.ax.set_ylabel(colorbar_title)
        cb.ax.tick_params(direction="in")

    # Set plot title and labels
    ax.set_title(title, fontweight="bold")
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_aspect("equal")

    # Style plot borders and ticks
    for spine in ax.spines.values():
        spine.set_linewidth(2.0)
    ax.tick_params(axis="both", which="major", width=2.0, length=3)


def vdfslice(
    meta: Vlsv,
    location: tuple | list,
    ax=None,
    limits: tuple = (float("-inf"), float("inf"), float("-inf"), float("inf")),
    verbose: bool = False,
    species: str = "proton",
    unit: AxisUnit = AxisUnit.SI,
    unitv: str = "km/s",
    vmin: float = float("-inf"),
    vmax: float = float("inf"),
    slicetype: str = None,
    vslicethick: float = 0.0,
    center: str = None,
    weight: str = "particle",
    flimit: float = -1.0,
    addcolorbar: bool = True,
    **kwargs,
):
    v1, v2, r1, r2, weights, strx, stry, str_title = prep_vdf(
        meta,
        location,
        species,
        unit,
        unitv,
        slicetype,
        vslicethick,
        center,
        weight,
        flimit,
        verbose,
    )

    import matplotlib
    import matplotlib.pyplot as plt

    if math.isinf(vmin):
        vmin = np.min(weights)
    if math.isinf(vmax):
        vmax = np.max(weights)

    if verbose:
        print(f"Active f range is {vmin}, {vmax}")

    if not ax:
        ax = plt.gca()

    norm = matplotlib.colors.LogNorm(vmin, vmax)

    h = ax.hist2d(v1, v2, bins=(r1, r2), weights=weights, norm=norm, shading="flat")

    ax.set_title(str_title, fontweight="bold")
    ax.set_xlabel(strx)
    ax.set_ylabel(stry)
    ax.set_aspect("equal")
    ax.grid(color="grey", linestyle="-")
    ax.tick_params(direction="in")

    if addcolorbar:
        cb = plt.colorbar(h[3], ax=ax, fraction=0.04, pad=0.02)
        cb.ax.tick_params(which="both", direction="in")
        cb_title = cb.ax.set_ylabel("f(v)")

    # TODO: Draw vector of magnetic field direction
    # if slicetype in ("xy", "xz", "yz"):

    return h[3]  # h[0] is 2D data, h[1] is x axis, h[2] is y axis


def prep_vdf(
    meta: Vlsv,
    location: tuple | list,
    species: str = "proton",
    unit: AxisUnit = AxisUnit.SI,
    unitv: str = "km/s",
    slicetype: str = None,
    vslicethick: float = 0.0,
    center: str = None,
    weight: str = "particle",
    flimit: float = -1.0,
    verbose: bool = False,
):
    ncells = meta.ncells

    if species in meta.meshes:
        vmesh = meta.meshes[species]
    else:
        raise ValueError(f"Unable to detect population {species}")

    if not slicetype in (None, "xy", "xz", "yz", "bperp", "bpar1", "bpar2"):
        raise ValueError(f"Unknown type {slicetype}")

    if unit == AxisUnit.EARTH:
        location = [loc * RE for loc in location]

    # Set unit conversion factor
    unitvfactor = 1e3 if unitv == "km/s" else 1.0

    # Get closest cell ID from input coordinates
    cidReq = meta.getcell(location)
    cidNearest = meta.getnearestcellwithvdf(cidReq)

    # Set normal direction
    if not slicetype:
        if ncells[1] == 1 and ncells[2] == 1:  # 1D, select xz
            slicetype = "xz"
        elif ncells[1] == 1:  # polar
            slicetype = "xz"
        elif ncells[2] == 1:  # ecliptic
            slicetype = "xy"
        else:
            slicetype = "xy"

    if slicetype in ("xy", "yz", "xz"):
        if slicetype == "xy":
            dir1, dir2, dir3 = 0, 1, 2
            ŝ = [0.0, 0.0, 1.0]
        elif slicetype == "xz":
            dir1, dir2, dir3 = 0, 2, 1
            ŝ = [0.0, 1.0, 0.0]
        elif slicetype == "yz":
            dir1, dir2, dir3 = 1, 2, 0
            ŝ = [1.0, 0.0, 0.0]
        v1size = vmesh.vblocks[dir1] * vmesh.vblock_size[dir1]
        v2size = vmesh.vblocks[dir2] * vmesh.vblock_size[dir2]

        v1min, v1max = vmesh.vmin[dir1], vmesh.vmax[dir1]
        v2min, v2max = vmesh.vmin[dir2], vmesh.vmax[dir2]
    elif slicetype in ("bperp", "bpar1", "bpar2"):
        # TODO: WIP
        pass

    if not math.isclose((v1max - v1min) / v1size, (v2max - v2min) / v2size):
        warnings.warn("Noncubic vgrid applied!")

    cellsize = (v1max - v1min) / v1size

    r1 = np.linspace(v1min / unitvfactor, v1max / unitvfactor, v1size + 1)
    r2 = np.linspace(v2min / unitvfactor, v2max / unitvfactor, v2size + 1)

    vcellids, vcellf = meta.read_vcells(cidNearest, species)

    V = meta.getvcellcoordinates(vcellids, species)

    if center:
        if center == "bulk":  # centered with bulk velocity
            if meta.has_variable("moments"):  # From a restart file
                Vcenter = meta.read_variable("restart_V", cidNearest)
            elif meta.has_variable(species * "/vg_v"):  # Vlasiator 5
                Vcenter = meta.read_variable(species * "/vg_v", cidNearest)
            elif meta.has_variable(species * "/V"):
                Vcenter = meta.read_variable(species * "/V", cidNearest)
            else:
                Vcenter = meta.read_variable("V", cidNearest)
        elif center == "peak":  # centered on highest VDF-value
            Vcenter = np.maximum(vcellf)

        V = np.array(
            [np.fromiter((v[i] - Vcenter for i in range(3)), dtype=float) for v in V]
        )

    # Set sparsity threshold
    if flimit < 0:
        if meta.has_variable(species + "/vg_effectivesparsitythreshold"):
            flimit = meta.readvariable(
                species + "/vg_effectivesparsitythreshold", cidNearest
            )
        elif meta.has_variable(species + "/EffectiveSparsityThreshold"):
            flimit = meta.read_variable(
                species + "/EffectiveSparsityThreshold", cidNearest
            )
        else:
            flimit = 1e-16

    # Drop velocity cells which are below the sparsity threshold
    findex_ = vcellf >= flimit
    fselect = vcellf[findex_]
    Vselect = V[findex_]

    if slicetype in ("xy", "yz", "xz"):
        v1select = np.array([v[dir1] for v in Vselect])
        v2select = np.array([v[dir2] for v in Vselect])
        vnormal = np.array([v[dir3] for v in Vselect])

        vec = ("vx", "vy", "vz")
        strx = vec[dir1]
        stry = vec[dir2]
    elif slicetype in ("bperp", "bpar1", "bpar2"):
        v1select = np.empty(len(Vselect), dtype=Vselect.dtype)
        v2select = np.empty(len(Vselect), dtype=Vselect.dtype)
        vnormal = np.empty(len(Vselect), dtype=Vselect.dtype)
        # TODO: WIP
        # for v1s, v2s, vn, vs in zip(v1select, v2select, vnormal, Vselect):
        #    v1s, v2s, vn = Rinv * Vselect

        # if slicetype == "bperp":
        #    strx = r"$v_{B \times V}$"
        #    stry = r"$v_{B \times (B \times V)}$"
        # elif slicetype == "bpar2":
        #    strx = r"$v_{B}$"
        #    stry = r"$v_{B \times V}$"
        # elif slicetype == "bpar1":
        #    strx = r"$v_{B \times (B \times V)}$"
        #    stry = r"$v_{B}$"

    unitstr = f" [{unitv}]"
    strx += unitstr
    stry += unitstr

    if vslicethick < 0:  # Set a proper thickness
        if any(
            i == 1.0 for i in ŝ
        ):  # Assure that the slice cut through at least 1 vcell
            vslicethick = cellsize
        else:  # Assume cubic vspace grid, add extra space
            vslicethick = cellsize * (math.sqrt(3) + 0.05)

    # Weights using particle flux or phase-space density
    fweight = (
        fselect * np.linalg.norm([v1select, v2select, vnormal])
        if weight == "flux"
        else fselect
    )

    # Select cells within the slice area
    if vslicethick > 0.0:
        ind_ = abs(vnormal) <= 0.5 * vslicethick
        v1, v2, fweight = v1select[ind_], v2select[ind_], fweight[ind_]
    else:
        v1, v2 = v1select, v2select

    v1 = [v / unitvfactor for v in v1]
    v2 = [v / unitvfactor for v in v2]

    str_title = f"t = {meta.time:4.1f}s"

    if verbose:
        print(f"Original coordinates : {location}")
        print(f"Original cell        : {meta.getcellcoordinates(cidReq)}")
        print(f"Nearest cell with VDF: {meta.getcellcoordinates(cidNearest)}")
        print(f"CellID: {cidNearest}")

        if center == "bulk":
            print(f"Transforming to plasma frame, travelling at speed {Vcenter}")
        elif center == "peak":
            print(f"Transforming to peak f-value frame, travelling at speed {Vcenter}")

        print(f"Using VDF threshold value of {flimit}.")

        if vslicethick > 0:
            print("Performing slice with a counting thickness of $vslicethick")
        else:
            print(f"Projecting total VDF to a single plane")

    return v1, v2, r1, r2, fweight, strx, stry, str_title


# Append plotting functions
Vlsv.plot = plot
Vlsv.pcolormesh = pcolormesh
Vlsv.contourf = contourf
Vlsv.contour = contour
Vlsv.streamplot = streamplot
Vlsv.vdfslice = vdfslice



"""
Plotting functionality API.

This provides supports for plotting VLSV data via Matplotlib.
"""

from pyvlasiator.plot.plot import (
    plot,
    pcolormesh,
    contour,
    contourf,
    streamplot,
    vdfslice,
)



from xml.etree import ElementTree
import numpy as np
import os
from collections import namedtuple
from pyvlasiator.vlsv.variables import units_predefined


class VMeshInfo:
    def __init__(
        self,
        vblocks: np.ndarray,
        vblock_size: np.ndarray,
        vmin: np.ndarray,
        vmax: np.ndarray,
        dv: np.ndarray,
    ) -> None:
        self.vblocks = vblocks
        self.vblock_size = vblock_size
        self.vmin = vmin
        self.vmax = vmax
        self.dv = dv
        self.cellwithVDF = np.empty(0, dtype=np.uint64)
        self.nblock_C = np.empty(0, dtype=np.int64)

        self.vblocks.flags.writeable = False
        self.vblock_size.flags.writeable = False
        self.vmin.flags.writeable = False
        self.vmax.flags.writeable = False
        self.dv.flags.writeable = False


"Variable information from the VLSV footer."
VarInfo = namedtuple(
    "VarInfo", ["unit", "unitLaTeX", "variableLaTeX", "unitConversion"]
)


class Vlsv:
    def __init__(self, filename: str):
        self.dir, self.name = os.path.split(filename)
        self.fid = open(filename, "rb")
        self.xmlroot = ElementTree.fromstring("<VLSV></VLSV>")
        self.celldict = {}
        self.maxamr = -1
        self.vg_indexes_on_fg = np.array([])  # SEE: map_vg_onto_fg(self)

        self._read_xml_footer()

        if self.has_parameter(name="time"):  # Vlasiator 5.0+
            self.time = self.read(name="time", tag="PARAMETER")
        elif self.has_parameter(name="t"):
            self.time = self.read(name="t", tag="PARAMETER")
        else:
            self.time = -1.0

        # Check if the file is using new or old VLSV format
        # Read parameters
        meshName = "SpatialGrid"
        bbox = self.read(tag="MESH_BBOX", mesh=meshName)
        if bbox is None:
            try:
                # Vlasiator 4- files where the mesh is defined with parameters
                self.ncells = np.array(
                    (
                        self.read_parameter("xcells_ini"),
                        self.read_parameter("ycells_ini"),
                        self.read_parameter("zcells_ini"),
                    ),
                    dtype=int,
                )
                self.block_size = np.array((1, 1, 1), dtype=int)
                self.coordmin = np.array(
                    (
                        self.read_parameter("xmin"),
                        self.read_parameter("ymin"),
                        self.read_parameter("zmin"),
                    ),
                    dtype=float,
                )
                self.coordmax = np.array(
                    (
                        self.read_parameter("xmax"),
                        self.read_parameter("ymax"),
                        self.read_parameter("zmax"),
                    ),
                    dtype=float,
                )
            except:  # dummy values
                self.ncells = np.array((1, 1, 1), dtype=int)
                self.block_size = np.array((1, 1, 1), dtype=int)
                self.coordmin = np.array((0.0, 0.0, 0.0), dtype=float)
                self.coordmax = np.array((1.0, 1.0, 1.0), dtype=float)

        else:
            # Vlasiator 5+ file
            nodeX = self.read(tag="MESH_NODE_CRDS_X", mesh=meshName)
            nodeY = self.read(tag="MESH_NODE_CRDS_Y", mesh=meshName)
            nodeZ = self.read(tag="MESH_NODE_CRDS_Z", mesh=meshName)
            self.ncells = np.fromiter((i for i in bbox[0:3]), dtype=int)
            self.block_size = np.fromiter((i for i in bbox[3:6]), dtype=int)
            self.coordmin = np.array((nodeX[0], nodeY[0], nodeZ[0]), dtype=float)
            self.coordmax = np.array((nodeX[-1], nodeY[-1], nodeZ[-1]), dtype=float)

        self.dcoord = np.fromiter(
            ((self.coordmax[i] - self.coordmin[i]) / self.ncells[i] for i in range(3)),
            dtype=float,
        )

        self.ncells.flags.writeable = False
        self.block_size.flags.writeable = False
        self.coordmin.flags.writeable = False
        self.coordmax.flags.writeable = False
        self.dcoord.flags.writeable = False

        self.meshes = {}

        # Iterate through the XML tree, find all populations
        self.species = []

        for child in self.xmlroot.findall("BLOCKIDS"):
            if "name" in child.attrib:
                popname = child.attrib["name"]
            else:
                popname = "avgs"

            if not popname in self.species:
                self.species.append(popname)

            bbox = self.read(tag="MESH_BBOX", mesh=popname)
            if bbox is None:
                if self.read_parameter("vxblocks_ini") is not None:
                    # Vlasiator 4- files where the mesh is defined with parameters
                    vblocks = np.array(
                        (
                            self.read_parameter("vxblocks_ini"),
                            self.read_parameter("vyblocks_ini"),
                            self.read_parameter("vzblocks_ini"),
                        ),
                        dtype=int,
                    )
                    vblock_size = np.array((4, 4, 4), dtype=int)
                    vmin = np.array(
                        (
                            self.read_parameter("vxmin"),
                            self.read_parameter("vymin"),
                            self.read_parameter("vzmin"),
                        ),
                        dtype=float,
                    )
                    vmax = np.array(
                        (
                            self.read_parameter("vxmax"),
                            self.read_parameter("vymax"),
                            self.read_parameter("vzmax"),
                        ),
                        dtype=float,
                    )

                else:  # no velocity space
                    vblocks = np.array((0, 0, 0), dtype=int)
                    vblock_size = np.array((4, 4, 4), dtype=int)
                    vmin = np.array((0.0, 0.0, 0.0), dtype=float)
                    vmax = np.array((0.0, 0.0, 0.0), dtype=float)
                    dv = np.array((1.0, 1.0, 1.0), dtype=float)

            else:  # Vlasiator 5+ file with bounding box
                nodeX = self.read(tag="MESH_NODE_CRDS_X", mesh=popname)
                nodeY = self.read(tag="MESH_NODE_CRDS_Y", mesh=popname)
                nodeZ = self.read(tag="MESH_NODE_CRDS_Z", mesh=popname)

                vblocks = np.array((*bbox[0:3],), dtype=int)
                vblock_size = np.array((*bbox[3:6],), dtype=int)
                vmin = np.array((nodeX[0], nodeY[0], nodeZ[0]), dtype=float)
                vmax = np.array((nodeX[-1], nodeY[-1], nodeZ[-1]), dtype=float)

            dv = np.fromiter(
                ((vmax[i] - vmin[i]) / vblocks[i] / vblock_size[i] for i in range(3)),
                dtype=float,
            )

            self.meshes[popname] = VMeshInfo(vblocks, vblock_size, vmin, vmax, dv)

            # Precipitation energy bins
            i = 0
            energybins = []
            binexists = True
            while binexists:
                binexists = self.has_parameter(
                    f"{popname}_PrecipitationCentreEnergy{i}"
                )
                if binexists:
                    binvalue = self.read_parameter(
                        f"{popname}_PrecipitationCentreEnergy{i}"
                    )
                    energybins.append(binvalue)
                i += 1
            if i > 1:
                self.precipitationenergybins[popname] = energybins

        self.variable = [
            node.attrib["name"] for node in self.xmlroot.findall("VARIABLE")
        ]

        cellid = self.read(mesh="SpatialGrid", name="CellID", tag="VARIABLE")
        self.cellindex = np.argsort(cellid)
        self.celldict = {cid: i for (i, cid) in enumerate(cellid)}

        self.maxamr = self.getmaxrefinement(cellid)

        self.nodecellwithVDF = self.xmlroot.findall("CELLSWITHBLOCKS")

        if len(self.nodecellwithVDF) == 0:
            self.hasvdf = False
        else:
            self.hasvdf = self.nodecellwithVDF[0].attrib["arraysize"] != "0"

    def __repr__(self) -> str:
        str = (
            f"File       : {self.name}\n"
            f"Time       : {self.time:.4f}\n"
            f"Dimension  : {self.ndims()}\n"
            f"Max AMR lvl: {self.maxamr}\n"
            f"Has VDF    : {self.hasvdf}\n"
            f"Variables  : {self.variable}\n"
        )
        return str

    def ndims(self) -> int:
        """Get the spatial dimension of data."""
        return sum(i > 1 for i in self.ncells)

    def _read_xml_footer(self) -> None:
        """Read the XML footer of the VLSV file."""
        fid = self.fid
        # first 8 bytes indicate endianness
        endianness_offset = 8
        fid.seek(endianness_offset)
        # offset of the XML file
        uint64_byte_amount = 8
        offset = int.from_bytes(fid.read(uint64_byte_amount), "little", signed=True)
        fid.seek(offset)
        xmlstring = fid.read()
        self.xmlroot = ElementTree.fromstring(xmlstring)

    def read(
        self,
        name: str = "",
        tag: str = "",
        mesh: str = "",
        cellids=-1,
    ) -> np.ndarray:
        """
        Read data of name, tag, and mesh from the vlsv file.

        This is the general reading function for all types of variables in VLSV files.

        Parameters
        ----------
        cellids : int or list of int
            If -1 then all data is read. If nonzero then only the vector for the specified
            cell id or cellids is read.
        Returns
        -------
        numpy.ndarray
        """
        if not tag and not name:
            raise ValueError("Must provide either tag or name")

        name = name.lower()

        fid = self.fid

        if "/" in name:
            popname, varname = name.split("/")
        else:
            popname, varname = "pop", name

        # TODO: add data reducers

        for child in self.xmlroot:
            if tag and child.tag != tag:
                continue
            if name and "name" in child.attrib and child.attrib["name"].lower() != name:
                continue
            if mesh and "mesh" in child.attrib and child.attrib["mesh"] != mesh:
                continue
            if child.tag == tag:
                vsize = int(child.attrib["vectorsize"])
                asize = int(child.attrib["arraysize"])
                dsize = int(child.attrib["datasize"])
                dtype = child.attrib["datatype"]
                variable_offset = int(child.text)

                # Select efficient method to read data based on number of cells
                if hasattr(cellids, "__len__"):
                    ncellids = len(cellids)
                    # Read multiple specified cells
                    # For reading a large amount of single cells, it'll be faster to
                    # read all data from the file and sort afterwards.
                    arraydata = []
                    if ncellids > 5000:
                        result_size = ncellids
                        read_size = asize
                        read_offsets = [0]
                    else:  # Read multiple cell ids one-by-one
                        result_size = ncellids
                        read_size = 1
                        read_offsets = [
                            self.celldict[cid] * dsize * vsize for cid in cellids
                        ]
                else:
                    if cellids < 0:  # all cells
                        result_size = asize
                        read_size = asize
                        read_offsets = [0]
                    else:  # parameter or single cell
                        result_size = 1
                        read_size = 1
                        read_offsets = [self.celldict[cellids] * dsize * vsize]

                for r_offset in read_offsets:
                    use_offset = int(variable_offset + r_offset)
                    fid.seek(use_offset)

                    if dtype == "float" and dsize == 4:
                        dtype = np.float32
                    elif dtype == "float" and dsize == 8:
                        dtype = np.float64
                    elif dtype == "int" and dsize == 4:
                        dtype = np.int32
                    elif dtype == "int" and dsize == 8:
                        dtype = np.int64
                    elif dtype == "uint" and dsize == 4:
                        dtype = np.uint32
                    elif dtype == "uint" and dsize == 8:
                        dtype = np.uint64

                    data = np.fromfile(fid, dtype, count=vsize * read_size)

                    if len(read_offsets) != 1:
                        arraydata.append(data)

                if len(read_offsets) == 1 and result_size < read_size:
                    # Many single cell IDs requested
                    # Pick the elements corresponding to the requested cells
                    for cid in cellids:
                        append_offset = self.celldict[cid] * vsize
                        arraydata.append(data[append_offset : append_offset + vsize])
                    data = np.squeeze(np.array(arraydata))
                elif len(read_offsets) != 1:
                    # Not so many single cell IDs requested
                    data = np.squeeze(np.array(arraydata))

                if vsize > 1:
                    data = data.reshape(result_size, vsize)

                if result_size == 1:
                    return data[0]
                else:
                    return data

        if name:
            raise NameError(
                name
                + "/"
                + tag
                + "/"
                + mesh
                + " not found in .vlsv file or in data reducers!"
            )

    def read_variable(
        self, name: str, cellids: int | list[int] | np.ndarray = -1, sorted: bool = True
    ) -> np.ndarray:
        """
        Read variables as numpy arrays from the open vlsv file.

        Parameters
        ----------
        cellids : int or list[int] or np.ndarray
            If -1 then all data is read. If nonzero then only the vector for the specified
            cell id or cellids is read.
        sorted : bool
            If the returned array is sorted by cell IDs. Only applied for full arrays.
        Returns
        -------
        numpy.ndarray
        """

        if self.has_variable(name) and name.startswith("fg_"):
            if not cellids == -1:
                raise ValueError("CellID requests not supported for FSgrid.")
            return self.read_fg_variable(name=name)

        if self.has_variable(name) and name.startswith("ig_"):
            if not cellids == -1:
                raise ValueError("CellID requests not supported for ionosphere.")
            return self.read_ionosphere_variable(name=name)

        raw = self.read(
            mesh="SpatialGrid",
            name=name,
            tag="VARIABLE",
            cellids=cellids,
        )
        if hasattr(cellids, "__len__"):  # part of cells requested
            return np.float32(raw)
        elif cellids == -1:
            if sorted:
                if raw.ndim == 1:
                    v = raw[self.cellindex]
                else:
                    v = raw[self.cellindex, :]
                if v.dtype == np.float64:  # 32-bit is enough for analysis
                    v = np.float32(v)
            else:
                v = raw
        else:
            v = raw

        return v

    def read_fg_variable(self, name: str):
        raw = self.read(
            mesh="fsgrid",
            name=name,
            tag="VARIABLE",
        )

        bbox = tuple(ncell * 2**self.maxamr for ncell in self.ncells)

        # Determine fsgrid domain decomposition
        nIORanks = self.read_parameter("numWritingRanks")  # Int32

        if raw.ndim > 1:
            v = np.empty((*bbox, raw.shape[-1]), dtype=np.float32)
        else:
            v = np.empty(bbox, dtype=np.float32)

        def getDomainDecomposition(globalsize, nproc: int) -> list[int]:
            """Obtain decomposition of this grid over the given number of processors.
            Reference: fsgrid.hpp
            """
            domainDecomp = (1, 1, 1)
            minValue = 1e20
            for i in range(1, min(nproc, globalsize[0]) + 1):
                iBox = max(globalsize[0] / i, 1.0)
                for j in range(1, min(nproc, globalsize[1]) + 1):
                    if i * j > nproc:
                        break
                    jBox = max(globalsize[1] / j, 1.0)
                    for k in range(1, min(nproc, globalsize[2]) + 1):
                        if i * j * k > nproc:
                            continue
                        kBox = max(globalsize[2] / k, 1.0)
                        v = (
                            10 * iBox * jBox * kBox
                            + ((jBox * kBox) if i > 1 else 0)
                            + ((iBox * kBox) if j > 1 else 0)
                            + ((iBox * jBox) if k > 1 else 0)
                        )
                        if i * j * k == nproc and v < minValue:
                            minValue = v
                            domainDecomp = (i, j, k)

            return domainDecomp

        def calcLocalStart(globalCells, nprocs: int, lcells: int) -> int:
            ncells = globalCells // nprocs
            remainder = globalCells % nprocs
            lstart = (
                lcells * (ncells + 1)
                if lcells < remainder
                else lcells * ncells + remainder
            )

            return lstart

        def calcLocalSize(globalCells, nprocs: int, lcells: int) -> int:
            ncells = globalCells // nprocs
            remainder = globalCells % nprocs
            lsize = ncells + 1 if lcells < remainder else ncells

            return lsize

        fgDecomposition = getDomainDecomposition(bbox, nIORanks)

        offsetnow = 0

        for i in range(nIORanks):
            xyz = (
                i // fgDecomposition[2] // fgDecomposition[1],
                i // fgDecomposition[2] % fgDecomposition[1],
                i % fgDecomposition[2],
            )

            lsize = tuple(
                map(
                    lambda i: calcLocalSize(bbox[i], fgDecomposition[i], xyz[i]),
                    range(0, 3),
                )
            )

            lstart = tuple(
                map(
                    lambda i: calcLocalStart(bbox[i], fgDecomposition[i], xyz[i]),
                    range(0, 3),
                )
            )

            offsetnext = offsetnow + np.prod(lsize)
            lend = tuple(st + si for st, si in zip(lstart, lsize))

            # Reorder data
            if raw.ndim > 1:
                v[lstart[0] : lend[0], lstart[1] : lend[1], lstart[2] : lend[2], :] = (
                    raw[offsetnow:offsetnext, :].reshape(
                        *lsize, raw.shape[-1], order="F"
                    )
                )
            else:
                v[lstart[0] : lend[0], lstart[1] : lend[1], lstart[2] : lend[2]] = raw[
                    offsetnow:offsetnext
                ].reshape(*lsize, order="F")

            offsetnow = offsetnext

        return v

    def read_variable_meta(self, var: str):
        unit, unitLaTeX, variableLaTeX, unitConversion = "", "", "", ""

        if var in units_predefined:
            unit, variableLaTeX, unitLaTeX = units_predefined[var]
        elif self.has_variable(var):  # For Vlasiator 5 files, MetaVLSV is included
            for child in self.xmlroot:
                if "name" in child.attrib and child.attrib["name"] == var:
                    if not "unit" in child.attrib:
                        break
                    else:
                        unit = child.attrib["unit"]
                        unitLaTeX = child.attrib["unitLaTeX"]
                        variableLaTeX = child.attrib["variableLaTeX"]
                        unitConversion = child.attrib["unitConversion"]

        return VarInfo(unit, unitLaTeX, variableLaTeX, unitConversion)

    def read_parameter(self, name: str):
        return self.read(name=name, tag="PARAMETER")

    def read_vcells(self, cellid: int, species: str = "proton"):
        """Read raw velocity block data.

        Parameters
        ----------
        cellid :
            Cell ID of the cell whose velocity blocks are read.
        species : str
            Population required.

        Returns
        -------
        numpy.ndarray
            A numpy array with block ids and their data.
        """

        fid = self.fid
        mesh = self.meshes[species]
        vblock_size = mesh.vblock_size

        self.init_cellswithVDF(species)

        # Check that cells have VDF stored
        try:
            cellWithVDFIndex = np.where(mesh.cellwithVDF == cellid)[0][0]
            nblocks = mesh.nblock_C[cellWithVDFIndex]
        except:
            raise ValueError(f"Cell ID {cellid} does not store VDF!")
        # Offset position to vcell storage
        offset_v = np.sum(mesh.nblock_C[0:cellWithVDFIndex], initial=0)

        # Read raw VDF
        for node in self.xmlroot.findall("BLOCKVARIABLE"):
            if node.attrib["name"] == species:
                dsize = int(node.attrib["datasize"])
                offset = int(node.text)
            break

        bsize = np.prod(vblock_size)
        fid.seek(offset_v * bsize * dsize + offset)
        T = np.float32 if dsize == 4 else np.float64
        data = np.fromfile(
            fid,
            dtype=T,
            count=bsize * nblocks,
        ).reshape(nblocks, bsize)

        # Read block IDs
        for node in self.xmlroot.findall("BLOCKIDS"):
            if node.attrib["name"] == species:
                dsize = int(node.attrib["datasize"])
                offset = int(node.text)
            break

        fid.seek(offset_v * dsize + offset)
        T = np.int32 if dsize == 4 else np.int64
        blockIDs = np.fromfile(fid, dtype=T, count=nblocks)

        # Velocity cell IDs and distributions (ordered by blocks)
        vcellids = np.empty(bsize * nblocks, dtype=np.int32)
        vcellf = np.empty(bsize * nblocks, dtype=np.float32)

        for i, bid in enumerate(blockIDs):
            for j in range(bsize):
                index_ = i * bsize + j
                vcellids[index_] = j + bsize * bid
                vcellf[index_] = data[i, j]

        return vcellids, vcellf

    def init_cellswithVDF(self, species: str = "proton") -> None:
        fid = self.fid
        mesh = self.meshes[species]
        if not np.any(mesh.cellwithVDF):
            for node in self.nodecellwithVDF:
                if node.attrib["name"] == species:
                    asize = int(node.attrib["arraysize"])
                    offset = int(node.text)
                    fid.seek(offset)
                    cellwithVDF = np.fromfile(fid, dtype=np.uint64, count=asize)
                    mesh.cellwithVDF = cellwithVDF
                    break

            for node in self.xmlroot.findall("BLOCKSPERCELL"):
                if node.attrib["name"] == species:
                    asize = int(node.attrib["arraysize"])
                    dsize = int(node.attrib["datasize"])
                    offset = int(node.text)
                    fid.seek(offset)
                    T = np.int32 if dsize == 4 else np.int64
                    nblock_C = np.fromfile(fid, dtype=T, count=asize).astype(np.int64)
                    mesh.nblock_C = nblock_C
                    break

            mesh.cellwithVDF = np.delete(mesh.cellwithVDF, np.where(mesh.nblock_C == 0))

    def _has_attribute(self, attribute: str, name: str) -> bool:
        """Check if a given attribute exists in the xml."""
        for child in self.xmlroot:
            if child.tag == attribute and "name" in child.attrib:
                if child.attrib["name"].lower() == name.lower():
                    return True
        return False

    def has_variable(self, name: str) -> bool:
        return self._has_attribute("VARIABLE", name)

    def has_parameter(self, name: str) -> bool:
        return self._has_attribute("PARAMETER", name)

    def getmaxrefinement(self, cellid: np.ndarray):
        """Get the maximum spatial refinement level."""
        ncell = np.prod(self.ncells)
        maxamr, cid = 0, ncell
        while cid < max(cellid):
            maxamr += 1
            cid += ncell * 8**maxamr

        return maxamr

    def getcell(self, loc: np.ndarray | tuple[int, ...] | list[int]):
        coordmin, coordmax = self.coordmin, self.coordmax
        dcoord = self.dcoord
        ncells = self.ncells
        celldict = self.celldict
        maxamr = self.maxamr

        for i in range(3):
            if not coordmin[i] < loc[i] < coordmax[i]:
                raise ValueError(f"{i} coordinate out of bound!")

        indices = np.fromiter(
            ((loc[i] - coordmin[i]) // dcoord[i] for i in range(3)), dtype=int
        )

        cid = (
            indices[0] + indices[1] * ncells[0] + indices[2] * ncells[0] * ncells[1] + 1
        )

        ncells_lowerlevel = 0
        ncell = np.prod(ncells)

        for ilevel in range(maxamr):
            if cid in celldict:
                break
            ncells_lowerlevel += (8**ilevel) * ncell
            ratio = 2 ** (ilevel + 1)
            indices = np.fromiter(
                (
                    np.floor((loc[i] - coordmin[i]) / dcoord[i] * ratio)
                    for i in range(3)
                ),
                dtype=int,
            )
            cid = (
                ncells_lowerlevel
                + indices[0]
                + ratio * ncells[0] * indices[1]
                + ratio**2 * ncells[0] * ncells[1] * indices[2]
                + 1
            )

        return cid

    def getvcellcoordinates(
        self, vcellids: np.ndarray, species: str = "proton"
    ) -> np.ndarray:
        mesh = self.meshes[species]
        vblocks = mesh.vblocks
        vblock_size = mesh.vblock_size
        dv = mesh.dv
        vmin = mesh.vmin

        bsize = np.prod(vblock_size)
        blockid = np.fromiter((cid // bsize for cid in vcellids), dtype=int)
        # Get block coordinates
        blockInd = [
            np.array(
                (
                    bid % vblocks[0],
                    bid // vblocks[0] % vblocks[1],
                    bid // (vblocks[0] * vblocks[1]),
                ),
                dtype=int,
            )
            for bid in blockid
        ]
        blockCoord = [
            np.array(
                (
                    bInd[0] * dv[0] * vblock_size[0] + vmin[0],
                    bInd[1] * dv[1] * vblock_size[1] + vmin[1],
                    bInd[2] * dv[2] * vblock_size[2] + vmin[2],
                ),
                dtype=float,
            )
            for bInd in blockInd
        ]
        # Get cell indices
        vcellblockids = np.fromiter((vid % bsize for vid in vcellids), dtype=int)
        cellidxyz = np.array(
            [
                np.array(
                    (
                        cid % vblock_size[0],
                        cid // vblock_size[0] % vblock_size[1],
                        cid // (vblock_size[0] * vblock_size[1]),
                    ),
                    dtype=int,
                )
                for cid in vcellblockids
            ]
        )
        # Get cell coordinates
        cellCoords = np.array(
            [
                np.fromiter(
                    (
                        blockCoord[i][j] + (cellidxyz[i][j] + 0.5) * dv[j]
                        for j in range(3)
                    ),
                    dtype=float,
                )
                for i in range(len(vcellids))
            ]
        )

        return cellCoords

    def getnearestcellwithvdf(self, id: int, species: str = "proton"):
        self.init_cellswithVDF(species)
        cells = self.meshes[species].cellwithVDF
        if not np.any(cells):
            raise ValueError(f"No distribution saved in {self.name}")
        coords_orig = self.getcellcoordinates(id)
        coords = [self.getcellcoordinates(cid) for cid in cells]
        min_ = np.argmin(np.sum(np.square(coords - coords_orig), axis=1))

        return cells[min_]

    def getcellcoordinates(self, cid: int):
        ncells = self.ncells
        coordmin, coordmax = self.coordmin, self.coordmax
        cid -= 1  # for easy divisions

        ncells_refmax = list(ncells)
        reflevel = 0
        subtraction = np.prod(ncells) * (2**reflevel) ** 3
        # sizes on the finest level
        while cid >= subtraction:
            cid -= subtraction
            reflevel += 1
            subtraction *= 8
            ncells_refmax[0] *= 2
            ncells_refmax[1] *= 2
            ncells_refmax[2] *= 2

        indices = np.array(
            (
                cid % ncells_refmax[0],
                cid // ncells_refmax[0] % ncells_refmax[1],
                cid // (ncells_refmax[0] * ncells_refmax[1]),
            ),
            dtype=int,
        )

        coords = np.fromiter(
            (
                coordmin[i]
                + (indices[i] + 0.5) * (coordmax[i] - coordmin[i]) / ncells_refmax[i]
                for i in range(3)
            ),
            dtype=float,
        )

        return coords

    def getslicecell(
        self, sliceoffset: float, dir: int, minCoord: float, maxCoord: float
    ):
        if not dir in (0, 1, 2):
            raise ValueError(f"Unknown slice direction {dir}")

        ncells, maxamr, celldict = self.ncells, self.maxamr, self.celldict
        nsize = ncells[dir]
        sliceratio = sliceoffset / (maxCoord - minCoord)
        if not (0.0 <= sliceratio <= 1.0):
            raise ValueError("slice plane index out of bound!")

        # Find the ids
        nlen = 0
        ncell = np.prod(ncells)
        # number of cells up to each refinement level
        lvlC = np.fromiter((ncell * 8**ilvl for ilvl in range(maxamr + 1)), dtype=int)
        lvlAccum = np.add.accumulate(lvlC)
        nStart = np.insert(lvlAccum, 0, 0)

        indexlist = np.empty(0, dtype=int)
        idlist = np.empty(0, dtype=int)

        cellidsorted = np.fromiter(celldict.keys(), dtype=int)
        cellidsorted.sort()

        for ilvl in range(maxamr + 1):
            nLow, nHigh = nStart[ilvl], nStart[ilvl + 1]
            idfirst_ = np.searchsorted(cellidsorted, nLow + 1)
            idlast_ = np.searchsorted(cellidsorted, nHigh, side="right")

            ids = cellidsorted[idfirst_:idlast_]

            ix, iy, iz = getindexes(ilvl, ncells[0], ncells[1], nLow, ids)

            if dir == 0:
                coords = ix
            elif dir == 1:
                coords = iy
            else:
                coords = iz

            # Find the cut plane index for each refinement level (0-based)
            depth = int(np.floor(sliceratio * nsize * 2**ilvl))
            # Find the needed elements to create the cut and save the results
            elements = coords == depth
            indexlist = np.append(indexlist, np.arange(nlen, nlen + len(ids))[elements])
            idlist = np.append(idlist, ids[elements])

            nlen += len(ids)

        return idlist, indexlist

    def refineslice(self, idlist: np.ndarray, data: np.ndarray, normal: int):
        ncells, maxamr = self.ncells, self.maxamr

        dims = _getdim2d(ncells, maxamr, normal)
        # meshgrid-like 2D input for matplotlib
        dpoints = np.empty((dims[1], dims[0]), dtype=data.dtype)

        # Create the plot grid
        ncell = np.prod(ncells)
        nHigh, nLow = ncell, 0

        for i in range(maxamr + 1):
            idfirst_ = np.searchsorted(idlist, nLow + 1)
            idlast_ = np.searchsorted(idlist, nHigh, side="right")

            ids = idlist[idfirst_:idlast_]
            d = data[idfirst_:idlast_]

            ix, iy, iz = getindexes(i, ncells[0], ncells[1], nLow, ids)

            # Get the correct coordinate values and the widths for the plot
            if normal == 0:
                a, b = iy, iz
            elif normal == 1:
                a, b = ix, iz
            elif normal == 2:
                a, b = ix, iy

            # Insert the data values into dpoints
            refineRatio = 2 ** (maxamr - i)
            iRange = range(refineRatio)
            X, Y = np.meshgrid(iRange, iRange, indexing="ij")
            coords = np.empty((len(a), 2 ** (2 * (maxamr - i)), 2), dtype=int)

            for ic, (ac, bc) in enumerate(zip(a, b)):
                for ir in range(2 ** (2 * (maxamr - i))):
                    index_ = np.unravel_index(ir, (refineRatio, refineRatio))
                    coords[ic, ir] = [
                        ac * refineRatio + X[index_],
                        bc * refineRatio + Y[index_],
                    ]

            for ic, dc in enumerate(d):
                for ir in range(2 ** (2 * (maxamr - i))):
                    dpoints[coords[ic, ir, 1], coords[ic, ir, 0]] = dc

            nLow = nHigh
            nHigh += ncell * 8 ** (i + 1)

        return dpoints


def _getdim2d(ncells: tuple, maxamr: int, normal: int):
    ratio = 2**maxamr
    if normal == 0:
        i1, i2 = 1, 2
    elif normal == 1:
        i1, i2 = 0, 2
    elif normal == 2:
        i1, i2 = 0, 1
    dims = (ncells[i1] * ratio, ncells[i2] * ratio)

    return dims


def getindexes(
    ilevel: int, xcells: int, ycells: int, nCellUptoLowerLvl: int, ids: np.ndarray
):
    ratio = 2**ilevel
    slicesize = xcells * ycells * ratio**2

    iz = (ids - nCellUptoLowerLvl - 1) // slicesize
    iy = np.zeros_like(iz)
    ix = np.zeros_like(iz)

    # number of ids up to the coordinate z in the refinement level ilevel
    idUpToZ = iz * slicesize + nCellUptoLowerLvl
    iy = (ids - idUpToZ - 1) // (xcells * ratio)
    ix = ids - idUpToZ - iy * xcells * ratio - 1

    return ix, iy, iz
